#include "SignatureImage.hpp"

#include <Magick++.h>
#include <iostream>
#include <sstream>
#include <sys/stat.h>

#ifdef MagickLibVersionText
#define MAGICK_MAJOR_VERSION (MagickLibVersion >> 8)
#endif

#ifdef MAGICK_MAJOR_VERSION
#define SET_COLOR(val) val << 8
#if MAGICK_MAJOR_VERSION == 7
#define Magick_RGBA(color)                                                     \
    Magick::Color(SET_COLOR(color.r), SET_COLOR(color.g), SET_COLOR(color.b),  \
                  SET_COLOR(color.alpha))
#else
#define Magick_RGBA(color)                                                     \
    Magick::Color(SET_COLOR(color.r), SET_COLOR(color.g), SET_COLOR(color.b),  \
                  color.alpha)
#endif
#endif

using namespace signimage;

struct TextContentInfo {
    std::string text;
    size_t fontSz;
    std::string fontFamily;
};

SignatureImage::SignatureImage(std::string title, CertificateInfo *certInfo)
    : m_title{title}, m_certInfo{certInfo} {

    m_signatureView = std::make_unique<SignatureView>();
    m_textView = std::make_unique<TextView>();
    m_logoView = std::make_unique<LogoView>();

    m_logo = nullptr;
    m_logoSize = 0;
}

SignatureImage::SignatureImage(std::string title, std::string logoFile,
                               CertificateInfo *certInfo)
    : SignatureImage(title, certInfo) {
    m_logoPath = logoFile;
}

SignatureImage::SignatureImage(std::string title, unsigned char *logoData,
                               size_t logoSize, CertificateInfo *certInfo)
    : SignatureImage(title, certInfo) {
    m_logoSize = logoSize;
    m_logo = logoData;
}

void SignatureImage::setBackgroundColor(RGBAColor rgba) {
    m_signatureView->bgColor = rgba;
}

void SignatureImage::setTextColor(RGBAColor rgba) {
    m_signatureView->textColor = rgba;
}

void SignatureImage::setBorderColor(RGBAColor rgba) {
    m_signatureView->borderColor = rgba;
}

void SignatureImage::setBorderRadius(BorderRadius radius) {
    if (radius.cornerWidth < 0) {
#ifndef NDEBUG
        std::cerr << "[DEBUG]: cornerWidth < 0" << std::endl;
#endif
        radius.cornerWidth = 0;
    }

    if (radius.cornerHeight < 0) {
#ifndef NDEBUG
        std::cerr << "[DEBUG]: cornerHeight < 0" << std::endl;
#endif
        radius.cornerHeight = 0;
    }

    m_signatureView->radius = radius;
}

void SignatureImage::setBorderWidth(size_t width) {
    if (width < 0) {
#ifndef NDEBUG
        std::cerr << "[DEBUG]: width < 0" << std::endl;
#endif
        width = 0;
    }

    m_signatureView->borderWidth = width;
}

void SignatureImage::setSignatureSize(Size size) {
    if (size.width < 0) {
#ifndef NDEBUG
        std::cerr << "[DEBUG]: the width of the signature image < 0"
                  << std::endl;
#endif
        size.width = 0;
    }

    if (size.height < 0) {
#ifndef NDEBUG
        std::cerr << "[DEBUG]: the height of the signature image < 0"
                  << std::endl;
#endif
        size.height = 0;
    }

    m_signatureView->size = size;
}

void SignatureImage::setLogoScale(std::string resolution) {
    m_logoView->imgResol = resolution;
}

void SignatureImage::setLogoPreserveRation(bool state) {
    m_logoView->logoPreserveRatio = state;
}

void SignatureImage::setLogoPos(Pos pos) {
    if (pos.x < 0 || pos.x > m_signatureView->size.width) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The logoPos.x has an incorrect value"
                  << std::endl;
#endif
        pos.x = 0;
    }

    if (pos.y < 0 || pos.y > m_signatureView->size.height) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The logoPos.y has an incorrect value"
                  << std::endl;
#endif
        pos.y = 0;
    }

    m_logoView->logoPos = {pos.x, pos.y};
}

void SignatureImage::setTitlePos(Pos pos) {
    if (pos.x < 0 || pos.x > m_signatureView->size.width) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The titlePos.x has an incorrect value"
                  << std::endl;
#endif
        pos.x = 0;
    }

    if (pos.y < 0 || pos.y > m_signatureView->size.height) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The titlePos.y has an incorrect value"
                  << std::endl;
#endif
        pos.y = 0;
    }

    m_textView->titlePos = {pos.x, pos.y};
}

void SignatureImage::setTitleAlignment(TextAlignment alignment) {
    m_textView->titleAlignment = alignment;
}

void SignatureImage::setContentAlignment(TextAlignment alignment) {
    m_textView->contentAlignment = alignment;
}

void SignatureImage::setCertSerialPos(Pos pos) {
    if (pos.x < 0 || pos.x > m_signatureView->size.width) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The certSerial.x has an incorrect value"
                  << std::endl;
#endif
        pos.x = 0;
    }

    if (pos.y < 0 || pos.y > m_signatureView->size.height) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The certSerial.y has an incorrect value"
                  << std::endl;
#endif
        pos.y = 0;
    }

    m_textView->certSerialPos = {pos.x, pos.y};
}

void SignatureImage::setSubjectPos(Pos pos) {
    if (pos.x < 0 || pos.x > m_signatureView->size.width) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The subjectPos.x has an incorrect value"
                  << std::endl;
#endif
        pos.x = 0;
    }

    if (pos.y < 0 || pos.y > m_signatureView->size.height) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The subjectPos.y has an incorrect value"
                  << std::endl;
#endif
        pos.y = 0;
    }

    m_textView->subjectPos = {pos.x, pos.y};
}

void SignatureImage::setTimeValidityPos(Pos pos) {
    if (pos.x < 0 || pos.x > m_signatureView->size.width) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The timeValidity.x has an incorrect value"
                  << std::endl;
#endif
        pos.x = 0;
    }

    if (pos.y < 0 || pos.y > m_signatureView->size.height) {
#ifndef NDEBUG
        std::cerr << "[DEBUG] The timeValidit.y has an incorrect value"
                  << std::endl;
#endif
        pos.y = 0;
    }

    m_textView->timeValidityPos = {pos.x, pos.y};
}

void SignatureImage::setTitleFontSize(size_t fontSize) {
    m_textView->titleFontSize = fontSize;
}

void SignatureImage::setFontSize(size_t fontSize) {
    m_textView->fontSize = fontSize;
}

void SignatureImage::setFontFamily(std::string font) {
    m_textView->fontFamily = font;
}

void SignatureImage::setRightMargin(double margin) {
    m_signatureView->rightMargin = margin;
}

double calcReqHeightField(TextContentInfo contentInfo, size_t fieldWidth) {
    Magick::Image textFieldImage(std::to_string(fieldWidth) + "x",
                                 Magick::Color("transparent"));

    textFieldImage.fontPointsize(contentInfo.fontSz);
    textFieldImage.fontFamily(contentInfo.fontFamily);
    textFieldImage.read("caption:" + contentInfo.text);

    return static_cast<double>(textFieldImage.rows());
}

void SignatureImage::updateFieldPos(TextFieldHeight textFieldH) {
    if (m_textView->certSerialPos.y <
        m_textView->titlePos.y + textFieldH.titleFieldHeight) {
        m_textView->certSerialPos.y = m_textView->titlePos.y +
                                      textFieldH.titleFieldHeight +
                                      (float)m_textView->fontSize / 2;
    }

    if (m_textView->subjectPos.y <
        m_textView->certSerialPos.y + textFieldH.certSerialFieldHeight) {
        m_textView->subjectPos.y = m_textView->certSerialPos.y +
                                   textFieldH.certSerialFieldHeight +
                                   (float)m_textView->fontSize / 2;
    }

    if (m_textView->timeValidityPos.y <
        m_textView->subjectPos.y + textFieldH.subjectFieldHeight) {
        m_textView->timeValidityPos.y = m_textView->subjectPos.y +
                                        textFieldH.subjectFieldHeight +
                                        (float)m_textView->fontSize / 2;
    }
}

Size SignatureImage::recalcSignatureSize() {
    auto width = m_signatureView->size.width -
                 (2 * m_signatureView->borderWidth +
                  m_signatureView->size.width * m_signatureView->rightMargin);
    auto height =
        m_signatureView->size.height - 2 * m_signatureView->borderWidth;

    auto calcTitleFieldWidth = width - m_textView->titlePos.x;
    auto calcCertSerialFieldWidth = width - m_textView->certSerialPos.x;
    auto calcSubjectFieldWidth = width - m_textView->subjectPos.x;
    auto calcTimeValidFieldWidth = width - m_textView->timeValidityPos.x;

    auto actualHeightTimeValidityField = height - m_textView->timeValidityPos.y;

    // the height of the field, taking into account the content
    auto reqHeightTitleField = calcReqHeightField(
        {m_title, m_textView->titleFontSize, m_textView->fontFamily},
        calcTitleFieldWidth);
    auto reqHeightCertSerialField = calcReqHeightField(
        {m_certInfo->cert_serial, m_textView->fontSize, m_textView->fontFamily},
        calcCertSerialFieldWidth);
    auto reqHeightSubjectField = calcReqHeightField(
        {m_certInfo->subject, m_textView->fontSize, m_textView->fontFamily},
        calcSubjectFieldWidth);
    auto reqHeightTimeValidField =
        calcReqHeightField({m_certInfo->timeValidity, m_textView->fontSize,
                            m_textView->fontFamily},
                           calcTimeValidFieldWidth);

    if (m_textView->certSerialPos.y >
            m_textView->titlePos.y + reqHeightTitleField &&
        m_textView->subjectPos.y >
            m_textView->certSerialPos.y + reqHeightCertSerialField &&
        m_textView->timeValidityPos.y >
            m_textView->subjectPos.y + reqHeightSubjectField &&
        m_textView->timeValidityPos.y + reqHeightTimeValidField <= height) {

        return m_signatureView->size;
    }

    updateFieldPos({reqHeightTitleField, reqHeightCertSerialField,
                    reqHeightSubjectField, reqHeightTimeValidField});

    // the minimum required height to fit the contents
    auto reqTotalHeight = m_textView->timeValidityPos.y +
                          reqHeightTimeValidField +
                          actualHeightTimeValidityField;

    return {m_signatureView->size.width, static_cast<double>(reqTotalHeight)};
}

struct TextInfo {
    std::string text;
    size_t fontSz;
    std::string fontFamily;
    Magick::GravityType graivity;
    Magick::Color textColor;
    Pos targetPos;
};

void createTextFieldImage(Magick::Image &image, size_t textFieldWidth,
                          size_t targetHeight, TextInfo &tInfo) {
    auto [textFieldPosX, textFieldPosY] = tInfo.targetPos;

    Magick::Image textFieldImage(std::to_string(textFieldWidth) + "x",
                                 Magick::Color("transparent"));

    textFieldImage.fontPointsize(tInfo.fontSz);
    textFieldImage.fontFamily(tInfo.fontFamily);
    textFieldImage.textGravity(tInfo.graivity);
    textFieldImage.fillColor(tInfo.textColor);
    textFieldImage.read("caption:" + tInfo.text);

    // remove the borders of text fields
    textFieldImage.border("0x0");

    image.composite(textFieldImage, textFieldPosX, textFieldPosY,
                    Magick::OverCompositeOp);
}

Resolution SignatureImage::getSignatureResolution() const {
    return m_signResolution;
}

std::vector<unsigned char> SignatureImage::getRawData() const {
    return m_rawData;
}

bool fileExists(std::string const &path) {
    struct stat st;

    return stat(path.c_str(), &st) == 0;
}

void SignatureImage::draw(ColorModel colorModel) {
    auto [width, height] = m_signatureView->size;

    if (width == 0 || height == 0) {
        return;
    }

    if (!m_logoPath.empty() && !fileExists(m_logoPath)) {
        return;
    }

    auto [newWidth, newHeight] = recalcSignatureSize();

    // calc the newWidth and newHeight without taking into account the newWidth
    // of the borders
    Magick::Image img(Magick::Geometry(newWidth, newHeight), "white");

    // the signature frame
    auto [rectPosX, rectPosY] = m_signatureView->size;

    img.fillColor(Magick_RGBA(m_signatureView->bgColor));
    img.strokeColor(Magick_RGBA(m_signatureView->borderColor));

    img.strokeWidth(m_signatureView->borderWidth);

    auto [rectCornerWidth, rectCornerHeight] = m_signatureView->radius;
    if (rectCornerWidth > 0 && rectCornerHeight > 0) {
        img.strokeAntiAlias(true);
        img.strokeLineJoin(Magick::RoundJoin);

        auto halfBorderWidth =
            static_cast<size_t>(floor(m_signatureView->borderWidth / 2));
        img.draw(Magick::DrawableRoundRectangle(
            halfBorderWidth, halfBorderWidth,
            img.baseColumns() - halfBorderWidth - 1,
            img.baseRows() - halfBorderWidth - 1, rectCornerWidth,
            rectCornerHeight));
    } else {
        img.draw(
            Magick::DrawableRectangle(0, 0, img.columns() - 1, img.rows() - 1));
    }

    // reset the stroke parameters
    img.strokeColor(Magick::Color("transparent"));
    img.strokeWidth(0);

    // The title of the signature
    img.fillColor(Magick::Color(""));
    TextInfo titleTextInfo = {
        m_title,
        m_textView->titleFontSize,
        m_textView->fontFamily,
        static_cast<Magick::GravityType>(m_textView->titleAlignment),
        Magick_RGBA(m_signatureView->textColor),
        m_textView->titlePos};
    auto calcTitleFieldWidth =
        (int)floor(newWidth -
                   (m_textView->titlePos.x + m_signatureView->borderWidth)) -
        m_signatureView->rightMargin * newWidth -
        m_signatureView->radius.cornerWidth;
    auto calcTitleTargetHeight =
        m_textView->certSerialPos.y - m_textView->titlePos.y;
    createTextFieldImage(img, calcTitleFieldWidth, calcTitleFieldWidth,
                         titleTextInfo);

    // The contents of the signature

    // The serial number of the certificate
    TextInfo certSerialTextInfo = {
        m_certInfo->cert_serial,
        m_textView->fontSize,
        m_textView->fontFamily,
        static_cast<Magick::GravityType>(m_textView->contentAlignment),
        Magick_RGBA(m_signatureView->textColor),
        m_textView->certSerialPos};
    auto calcCertSerialFieldWidth =
        (int)floor(newWidth - (m_textView->certSerialPos.x +
                               m_signatureView->borderWidth)) -
        m_signatureView->rightMargin * newWidth;
    auto calcCertSerialTargetHeight =
        m_textView->subjectPos.y - m_textView->certSerialPos.y;
    createTextFieldImage(img, calcCertSerialFieldWidth,
                         calcCertSerialTargetHeight, certSerialTextInfo);

    // The subject of the certificate
    TextInfo subjectTextInfo = {
        m_certInfo->subject,
        m_textView->fontSize,
        m_textView->fontFamily,
        static_cast<Magick::GravityType>(m_textView->contentAlignment),
        Magick_RGBA(m_signatureView->textColor),
        m_textView->subjectPos};
    auto calcSubjectTextFieldWidth =
        (int)floor(newWidth -
                   (m_textView->subjectPos.x + m_signatureView->borderWidth)) -
        m_signatureView->rightMargin * newWidth;
    auto calcSubjectTargetHeight =
        m_textView->timeValidityPos.y - m_textView->subjectPos.y;
    createTextFieldImage(img, calcSubjectTextFieldWidth,
                         calcSubjectTargetHeight, subjectTextInfo);

    // The validity time of the certificate
    TextInfo timeValidityTextInfo = {
        m_certInfo->timeValidity,
        m_textView->fontSize,
        m_textView->fontFamily,
        static_cast<Magick::GravityType>(m_textView->contentAlignment),
        Magick_RGBA(m_signatureView->textColor),
        m_textView->timeValidityPos};
    auto calcTimeValidTextFieldWidth =
        (int)floor(newWidth - (m_textView->timeValidityPos.x +
                               m_signatureView->borderWidth)) -
        m_signatureView->rightMargin * newWidth -
        m_signatureView->radius.cornerWidth;
    auto calcTimeValidTargetHeight = newHeight - m_textView->timeValidityPos.y;
    createTextFieldImage(img, calcTimeValidTextFieldWidth,
                         calcTimeValidTargetHeight, timeValidityTextInfo);

    // Reading the logo
    if (m_logo || !m_logoPath.empty()) {
        Magick::Image logo;
        if (m_logo) {
            try {
                Magick::Blob blobLogo(m_logo, m_logoSize);
                logo.read(blobLogo);
            } catch (Magick::Exception &err) {
#ifdef NDEBUG
                std::cerr << "[ERROR]: Invalid logo data!" << std::endl;
#endif
                std::cerr << "[ERROR]: Caught exception: " << err.what()
                          << std::endl;
            }
        } else if (!m_logoPath.empty()) {
            try {
                logo.read(m_logoPath);
            } catch (Magick::Exception &err) {
                std::cerr << "[ERROR]: Caught exception: " << err.what()
                          << std::endl;
            }
        }

        if (!logo.isValid()) {
#ifndef NDEBUG
            std::cerr << "[ERROR]: The logo data is invalid!" << std::endl;

            return;
#endif
        }

        auto resolution = m_logoView->imgResol;

        if (!resolution.empty()) {
#ifndef NDEBUG
            std::cerr << "[DEBUG]: The target size of the logo image: " +
                             resolution
                      << std::endl;
#endif
            if (m_logoView->logoPreserveRatio) {
                resolution = getRescaleLogoImage(resolution, logo.size());
            }

            resolution += "!";

            try {
                logo.resize(resolution);
            } catch (Magick::Exception &err) {
                std::cerr << "[ERROR]: Caught exception: " << err.what()
                          << std::endl;
            }

#ifndef NDEBUG
            std::cerr << "[DEBUG]: The actual size of the logo image: "
                      << logo.size().width() << "x" << logo.size().height()
                      << std::endl;
#endif
        }

        auto [logoPosX, logoPosY] = m_logoView->logoPos;

        // Align the stamp logo in the center if its actual width is less
        // than the specified width
        if (m_logoView->logoPreserveRatio) {
            centerHLogoImagePos(logoPosX, logo.size().width());
#ifndef NDEBUG
            std::cerr << "[DEBUG]: The position of the stamp logo has been "
                         "changes to "
                      << logoPosX << std::endl;
#endif
        }

        try {
            img.composite(logo, logoPosX, logoPosY, Magick::OverCompositeOp);
        } catch (Magick::Exception &err) {
            std::cerr << "[ERROR]: Caught exception: " << err.what()
                      << std::endl;
        }
    }

    // get raw rgba data
    auto signWidth = img.columns();
    auto signHeight = img.rows();

    m_signResolution = {signWidth, signHeight};

    try {
        if (colorModel == RGBA) {
            m_rawData.resize(signWidth * signHeight * 4);

            img.write(0, 0, signWidth, signHeight, "RGBA",
                      MagickCore::CharPixel, m_rawData.data());
        } else {
            m_rawData.resize(signWidth * signHeight * 3);

            img.write(0, 0, signWidth, signHeight, "RGB", MagickCore::CharPixel,
                      m_rawData.data());
        }
    } catch (Magick::Exception &err) {
        std::cerr << "[ERROR]: Caught exception: " << err.what() << std::endl;
    }
}

bool SignatureImage::isValidSizeValue(size_t &value) const {
    return value != std::numeric_limits<size_t>::max();
}

std::string
SignatureImage::getRescaleLogoImage(std::string const &resol,
                                    std::string const &sourceResol) const {
    auto targetLogoSize = imgSizeFromString(resol);

    if (isValidSizeValue(targetLogoSize.x) &&
        isValidSizeValue(targetLogoSize.y)) {
        auto sourceLogoSize = imgSizeFromString(sourceResol);
        auto resultLogoSize = scaleImage(sourceLogoSize, targetLogoSize);

        return imgSizeToString(resultLogoSize);
    }

    return {};
}

void SignatureImage::centerHLogoImagePos(double &logoPosX,
                                         size_t const &logoWidth) const {
    auto logoTargetWidth = imgSizeFromString(m_logoView->imgResol).x;

    if (!isValidSizeValue(logoTargetWidth) || logoWidth == logoTargetWidth) {
        return;
    }

    auto resultXPos = floor((logoPosX + logoTargetWidth - logoWidth) / 2);

    logoPosX = resultXPos >= 0 ? resultXPos : m_logoView->logoPos.x;
}

Resolution SignatureImage::imgSizeFromString(std::string const &str) const {
    auto delimPos = str.find("x");

    if (delimPos == std::string::npos) {
        return Resolution({std::numeric_limits<size_t>::max(),
                           std::numeric_limits<size_t>::max()});
    }

    auto targetLogoWidthStr = str.substr(0, delimPos);
    auto targetLogoHeightStr = str.substr(delimPos + 1);

    return Resolution({stringToImgSize(targetLogoWidthStr),
                       stringToImgSize(targetLogoHeightStr)});
}

size_t SignatureImage::stringToImgSize(const std::string str) const {
    size_t result = 0;
    std::istringstream ss(str);
    ss >> result;

    if (ss.fail()) {
        return std::numeric_limits<size_t>::max();
    }

    return result;
}

std::string SignatureImage::imgSizeToString(Resolution resol) const {
    std::stringstream ss;
    ss << resol.x << "x" << resol.y;

    if (ss.fail()) {
        return {};
    }

    return ss.str();
}

Resolution SignatureImage::scaleImage(Resolution imgSize,
                                      Resolution targetSize) const {
    double scaleFactor = 0;

    // scale along the larger side of the logo image
    if (imgSize.x >= imgSize.y) {
        scaleFactor = static_cast<double>(targetSize.x) / imgSize.x;
    } else {
        scaleFactor = static_cast<double>(targetSize.y) / imgSize.y;
    }

    auto resultWidth = imgSize.x * scaleFactor;
    auto resultHeight = imgSize.y * scaleFactor;

#ifndef NDEBUG
    std::cerr << "[DEBUG]: The width of the logo after scaling: " << resultWidth
              << std::endl;
    std::cerr << "[DEBUG]: The height of the logo after scaling: "
              << resultHeight << std::endl;
#endif

    Resolution resultSize = {static_cast<size_t>(floor(resultWidth)),
                             static_cast<size_t>(floor(resultHeight))};

    return resultSize;
}
